"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const gherkin_1 = require("@cucumber/gherkin");
const messages_1 = require("@cucumber/messages");
const prettier_1 = require("prettier");
const GherkinAST_1 = require("./GherkinAST");
const DEFAULT_ESCAPE_BACKSLASH = false;
// taken from https://github.com/cucumber/gherkin-javascript/blob/e25a1be3b21133c7a92eb7735997c6e774406226/src/GherkinClassicTokenMatcher.ts#L11
const LANGUAGE_PATTERN = /^\s*#\s*language\s*:\s*([a-zA-Z\-_]+)\s*$/;
const { literalline, hardline, join, group, trim, indent, line } = prettier_1.doc.builders;
const { hasNewline, isPreviousLineEmpty, makeString } = prettier_1.util;
const languages = [
    {
        extensions: ['.feature'],
        name: 'Gherkin',
        parsers: ['gherkin'],
    },
];
function printHardline() {
    // console.log(new Error().stack?.split('\n')[2]);
    return hardline;
}
function printTwoHardlines() {
    // console.log(new Error().stack?.split('\n')[2]);
    return [hardline, hardline];
}
const LINEBREAK_MATCHER = /\r\n|[\r\n\u2028\u2029]/;
/**
 * escaple line breaks in a string
 */
function escapeMultilineString(str) {
    return str.replace(new RegExp(LINEBREAK_MATCHER, 'g'), '\\n');
}
function assertNodeHasLocation(node) {
    if (!('location' in node)) {
        throw new Error('no location for node ' + node.constructor.name);
    }
}
let textColumnWidth = [];
function generateColumnSizes(text) {
    let columnSizes = [];
    let index = 0;
    let currentLine = 1;
    while ((index = prettier_1.util.skipEverythingButNewLine(text, index)) !== false) {
        columnSizes.push(index + 1);
        index++;
        currentLine++;
    }
    // console.log(columnSizes);
    textColumnWidth = columnSizes;
}
const gherkinParser = {
    parse: (text, parsers, options) => {
        var _a;
        const uuidFn = messages_1.IdGenerator.uuid();
        const builder = new gherkin_1.AstBuilder(uuidFn);
        const matcher = new gherkin_1.GherkinClassicTokenMatcher(); // or GherkinInMarkdownTokenMatcher()
        const parser = new gherkin_1.Parser(builder, matcher);
        const document = parser.parse(text);
        generateColumnSizes(text);
        // console.log({
        //   textColumnWidth,
        //   // document: JSON.stringify(document, null, 2),
        // });
        // console.log(
        //   util.getStringWidth(text),
        //   util.skipEverythingButNewLine(text, 18)
        // );
        return new GherkinAST_1.TypedGherkinDocument(Object.assign({}, document), {
            escapeBackslashes: (_a = options.escapeBackslashes) !== null && _a !== void 0 ? _a : DEFAULT_ESCAPE_BACKSLASH,
        });
    },
    locStart: (node) => {
        // return 0;
        // if (!(node instanceof TypedComment)) {
        //   throw new Error('locStart: not a comment');
        // }
        var _a;
        // console.log('locStart', node);
        assertNodeHasLocation(node);
        // sum all column size until the current line
        let index = node.location.line > 1 ? textColumnWidth[node.location.line - 2] : 0;
        index += ((_a = node.location.column) !== null && _a !== void 0 ? _a : 1) - 1;
        return index;
        // if (node instanceof TypedComment) {
        //   return 0;
        // }
        // if (node.location) {
        //   return node.location.line * 1000000 + node.location.column;
        // }
        // return 0;
        // return node.location.column ?? 0;
    },
    locEnd: (node) => {
        // return 0; // TMP ignore comments
        // if (!(node instanceof TypedComment)) {
        //   throw new Error('locEnd: not a comment');
        // }
        // console.log('locEnd', node, '\n');
        if (node instanceof GherkinAST_1.TypedComment) {
            return gherkinParser.locStart(node) + node.text.trim().length;
        }
        if (!(node instanceof GherkinAST_1.TypedComment)) {
            console.log({ method: 'locEnd', nodeName: node.constructor.name });
        }
        if (node instanceof GherkinAST_1.TypedFeature) {
            return (gherkinParser.locStart(node) +
                node.keyword.length +
                node.name.length +
                2);
        }
        // if (!(node instanceof TypedComment)) {
        //   console.log(node, gherkinParser.locStart(node));
        // }
        // if (node.location) {
        //   return node.location.line * 1000000 + node.location.column + 1;
        // }
        // console.log('locEnd', node);
        return gherkinParser.locStart(node) + 1;
        // return (node.location.column ?? 0) + node.text.length;
    },
    astFormat: 'gherkin-ast',
};
function printTags(path, node) {
    // console.log(node);
    return [
        // @ts-expect-error TODO path should be recognized as an AstPath<GherkinNode & { tags: readonly TypedTag[] }>>
        join(printHardline(), path.map(gherkinAstPrinter.print, 'tags')),
        node.tags.length > 0 ? printHardline() : '',
    ];
}
function printDescription(path, node, hardlineIfDescription) {
    // console.log(node);
    if (!node.description) {
        return hardlineIfDescription ? printHardline() : '';
    }
    return [
        join(printHardline(), node.description.split('\n').map((s) => s.trim())),
        printHardline(),
        printHardline(),
    ];
}
/**
 * This method will try to find the the location of the
 */
function findNodeForCommentInAST(ast, comment) {
    const { line, column } = comment.location;
    if ((0, GherkinAST_1.isWithLocation)(ast) && ast.location.line > line) {
        return ast;
    }
    if ((0, GherkinAST_1.isHasChild)(ast)) {
        const { child } = ast;
        if (child) {
            const node = findNodeForCommentInAST(child, comment);
            if (node) {
                return node;
            }
        }
    }
    if ((0, GherkinAST_1.isHasChildren)(ast)) {
        const { children } = ast;
        for (const child of children) {
            const node = findNodeForCommentInAST(child, comment);
            if (node) {
                return node;
            }
        }
    }
    return null;
    return null;
}
function stepNeedsHardline(node, isFirstStep) {
    return (!isFirstStep &&
        node.keywordType &&
        [messages_1.StepKeywordType.CONTEXT, messages_1.StepKeywordType.ACTION].includes(node.keywordType));
}
const gherkinAstPrinter = {
    printComment: (path, options) => {
        const node = path.getValue();
        // console.log(node);
        if (!(node instanceof GherkinAST_1.TypedComment)) {
            throw new Error('printComment: not a comment');
        }
        const parentNode = path.getParentNode();
        const parentNodeIsFirstStep = path.stack[path.stack.length - 6] === 0;
        // if the comment follows a `Given` step, then the comment should have a leading blank line
        // but no blank line between the comment and the step
        if (parentNode instanceof GherkinAST_1.TypedStep &&
            stepNeedsHardline(parentNode, parentNodeIsFirstStep)) {
            return [printHardline(), node.text.trim()];
        }
        return node.text.trim();
    },
    // canAttachComment(node): boolean {
    //   console.log('canAttachComment', node);
    //   return true;
    // },
    isBlockComment(node) {
        //   // console.log('isBlockComment', node);
        return false; // block comments are not supported by gherkin for now. See https://cucumber.io/docs/gherkin/reference/
    },
    handleComments: {
        ownLine: (commentNode, text, options, ast, isLastComment) => {
            const node = findNodeForCommentInAST(ast, commentNode);
            if (node) {
                prettier_1.util.addLeadingComment(node, commentNode);
                return true;
            }
            else {
                // we are on the latest lines of the file, and they all contains comments, so attach comment after the ast
                const isRemainingLinesComments = text
                    .split('\n')
                    .slice(commentNode.location.line)
                    .filter((row) => row.match(/^\s#/)).length === 0;
                if (isRemainingLinesComments) {
                    prettier_1.util.addTrailingComment(ast, commentNode);
                    return true;
                }
            }
            // console.log({ ownLine: commentNode, ast });
            return false;
        },
        endOfLine: (commentNode, text, options, ast, isLastComment) => {
            const node = findNodeForCommentInAST(ast, commentNode);
            if (node) {
                prettier_1.util.addLeadingComment(node, commentNode);
                return true;
            }
            // console.log({ ownLine: commentNode, ast });
            return false;
        },
        remaining: (commentNode, text, options, ast, isLastComment) => {
            if (ast instanceof GherkinAST_1.TypedGherkinDocument && !ast.feature) {
                // special case where the document only contains comments : let's attach the comment to the document directly
                prettier_1.util.addLeadingComment(ast, commentNode);
                return true;
            }
            return false;
        },
    },
    print: (path, options, print) => {
        const node = path.getValue();
        // console.log({ node, isDocument: node instanceof TypedGherkinDocument });
        if (node instanceof GherkinAST_1.TypedGherkinDocument) {
            if (node.feature) {
                // @ts-expect-error TODO  path should be recognized as an AstPath<TypedGherkinDocument>
                return [path.call(print, 'feature'), printHardline()];
            }
            else {
                // return empty sting if there is no feature (it can contain only comments)
                return '';
            }
        }
        else if (node instanceof GherkinAST_1.TypedFeature || node instanceof GherkinAST_1.TypedRule) {
            const hasLanguageInSourceFile = !!options.originalText.match(new RegExp(LANGUAGE_PATTERN, 'm'));
            return [
                node instanceof GherkinAST_1.TypedFeature && hasLanguageInSourceFile
                    ? ['# language: ' + node.language, printHardline()]
                    : '',
                printTags(path, node),
                `${node.keyword}: ${node.name}`,
                indent([
                    printHardline(),
                    printDescription(path, node, true),
                    // @ts-expect-error TODO path should be recognized as an AstPath<TypedFeature>
                    join(printTwoHardlines(), path.map(print, 'children')),
                ]),
            ];
        }
        else if (node instanceof GherkinAST_1.TypedFeatureChild ||
            node instanceof GherkinAST_1.TypedRuleChild) {
            if (node.scenario) {
                // @ts-expect-error TODO  path should be recognized as an AstPath<TypedFeatureChild>
                return path.call(print, 'scenario');
            }
            else if (node.background) {
                // @ts-expect-error TODO  path should be recognized as an AstPath<TypedFeatureChild>
                return path.call(print, 'background');
            }
            else if (node instanceof GherkinAST_1.TypedFeatureChild && node.rule) {
                // @ts-expect-error TODO  path should be recognized as an AstPath<TypedFeatureChild>
                return path.call(print, 'rule');
            }
            else {
                console.log(node);
                throw new Error(`unhandled case where ${node instanceof GherkinAST_1.TypedFeatureChild
                    ? 'TypedFeatureChild'
                    : 'TypedRuleChild'} has no scenario`);
            }
        }
        else if (node instanceof GherkinAST_1.TypedTag) {
            return node.name;
        }
        else if (node instanceof GherkinAST_1.TypedBackground) {
            // console.log(node.steps);
            return [
                `${node.keyword}: ${node.name}`,
                node.description || node.steps.length > 0
                    ? indent([
                        printHardline(),
                        printDescription(path, node, false),
                        // @ts-expect-error TODO  path should be recognized as an AstPath<TypedBackground>
                        join(printHardline(), path.map(print, 'steps')),
                    ])
                    : '',
            ];
        }
        else if (node instanceof GherkinAST_1.TypedScenario) {
            // console.log(node);
            return [
                printTags(path, node),
                `${node.keyword}: ${node.name}`,
                indent([
                    printHardline(),
                    printDescription(path, node, false),
                    // @ts-expect-error TODO path should be recognized as an AstPath<TypedScenario>
                    join(printHardline(), path.map(print, 'steps')),
                    node.steps.length > 0 && node.examples.length > 0
                        ? printTwoHardlines()
                        : '',
                    // @ts-expect-error TODO path should be recognized as an AstPath<TypedScenario>
                    join(printTwoHardlines(), path.map(print, 'examples')),
                ]),
            ];
        }
        else if (node instanceof GherkinAST_1.TypedStep) {
            // console.log(node);
            return [
                // if the step has comment, the hardline will be handled by the comment printer
                stepNeedsHardline(node, path.stack[path.stack.length - 2] === 0) &&
                    // @ts-expect-error comments are injected by prettier directly
                    !node.comments
                    ? printHardline()
                    : '',
                `${node.keyword}${node.text.trim()}`,
                node.docString
                    ? // @ts-expect-error TODO path should be recognized as an AstPath<TypedStep>
                        indent([printHardline(), path.call(print, 'docString')])
                    : '',
                node.dataTable // @ts-expect-error TODO path should be recognized as an AstPath<TypedStep>
                    ? indent([printHardline(), path.call(print, 'dataTable')])
                    : '',
            ];
        }
        else if (node instanceof GherkinAST_1.TypedDocString) {
            // console.log({ node });
            // if the content contains the delimiter, the parser will unescape it, so we need to escape it again
            const escapeDelimiter = (content) => {
                return content.replace(new RegExp(`(${node.delimiter})`, 'g'), (match) => {
                    return match
                        .split('')
                        .map((c) => `\\${c}`)
                        .join('');
                });
            };
            return join(printHardline(), [
                [node.delimiter, node.mediaType || ''],
                // split the string on newlines to preserve the indentation
                ...escapeDelimiter(node.content).split('\n'),
                node.delimiter,
            ]);
        }
        else if (node instanceof GherkinAST_1.TypedExamples) {
            // console.log({ node, columnSizes: node.columnSizes });
            return [
                printTags(path, node),
                `${node.keyword}: ${node.name}`,
                indent([
                    printHardline(),
                    printDescription(path, node, false),
                    join(printHardline(), [
                        // @ts-expect-error TODO path should be recognized as an AstPath<TypedExamples>
                        node.tableHeader && path.call(print, 'tableHeader'),
                        // @ts-expect-error TODO path should be recognized as an AstPath<TypedExamples>
                        ...path.map(print, 'tableBody'),
                    ].filter((d) => !!d)),
                ]),
            ];
        }
        else if (node instanceof GherkinAST_1.TypedDataTable) {
            // console.log({ node, columnSizes: node.columnSizes });
            return join(printHardline(), [
                // @ts-expect-error TODO path should be recognized as an AstPath<TypedDataTable>
                ...path.map(print, 'rows'),
            ]);
        }
        else if (node instanceof GherkinAST_1.TypedTableRow) {
            // console.log(node);
            return [
                '| ',
                // @ts-expect-error TODO path should be recognized as an AstPath<TypedTableRow>
                join(' | ', path.map(print, 'cells')),
                ' |',
            ];
        }
        else if (node instanceof GherkinAST_1.TypedTableCell) {
            // console.log(node);
            return escapeMultilineString(node.value.padEnd(node.displaySize));
        }
        else {
            console.error('Unhandled node type', node);
            return '';
        }
    },
    embed(path, print, textToDoc, options) {
        const node = path.getValue();
        if (node instanceof GherkinAST_1.TypedDocString) {
            const { content, mediaType } = node;
            if (mediaType === 'xml') {
                return [
                    node.delimiter,
                    mediaType,
                    printHardline(),
                    textToDoc(content, Object.assign(Object.assign({}, options), { parser: 'html' })),
                    node.delimiter,
                ];
            }
            try {
                JSON.parse(content);
                return [
                    node.delimiter,
                    mediaType !== null && mediaType !== void 0 ? mediaType : '',
                    printHardline(),
                    textToDoc(content, Object.assign(Object.assign({}, options), { parser: 'json' })),
                    node.delimiter,
                ];
            }
            catch (e) {
                // igonre non-JSON content
            }
        }
        return null;
    },
};
const plugin = {
    languages,
    parsers: {
        gherkin: gherkinParser,
    },
    printers: {
        'gherkin-ast': gherkinAstPrinter,
    },
    defaultOptions: {
        printWidth: 120,
        tabWidth: 4,
    },
    options: {
        escapeBackslashes: {
            type: 'boolean',
            default: DEFAULT_ESCAPE_BACKSLASH,
            description: 'Escape backslashes in strings',
            oppositeDescription: 'Do not escape backslashes in strings',
            since: '1.0.0',
            category: 'Format',
        },
    },
};
module.exports = plugin;
